"""A modification workflow that updates the :term:`LAG` interfaces that are part of an existing IP trunk."""

import json
from typing import Annotated, TypeVar
from uuid import UUID, uuid4

from annotated_types import Len
from orchestrator.forms import FormPage
from orchestrator.targets import Target
from orchestrator.types import FormGenerator, State, UUIDstr
from orchestrator.utils.json import json_dumps
from orchestrator.workflow import StepList, begin, conditional, done, step, workflow
from orchestrator.workflows.steps import resync, store_process_subscription, unsync
from orchestrator.workflows.utils import wrap_modify_initial_input_form
from pydantic import AfterValidator, ConfigDict, field_validator
from pydantic_forms.validators import Label, ReadOnlyField, validate_unique_list

from gso.products.product_blocks.iptrunk import (
    IptrunkInterfaceBlock,
    IptrunkSideBlock,
    IptrunkType,
    PhysicalPortCapacity,
)
from gso.products.product_types.iptrunk import Iptrunk
from gso.services.lso_client import execute_playbook, lso_interaction
from gso.services.netbox_client import NetboxClient
from gso.utils.helpers import (
    LAGMember,
    available_interfaces_choices,
    available_interfaces_choices_including_current_members,
    get_router_vendor,
    validate_interface_name_list,
    validate_iptrunk_unique_interface,
    validate_tt_number,
)
from gso.utils.shared_enums import IPv4AddressType, IPv6AddressType, Vendor
from gso.workflows.iptrunk.migrate_iptrunk import check_ip_trunk_optical_levels_pre
from gso.workflows.iptrunk.validate_iptrunk import check_ip_trunk_isis

T = TypeVar("T", bound=LAGMember)


def initialize_ae_members(
    subscription: Iptrunk, initial_user_input: dict, side_index: int
) -> Annotated[list[LAGMember], ""]:
    """Initialize the list of AE members."""
    router = subscription.iptrunk.iptrunk_sides[side_index].iptrunk_side_node
    router_vendor = get_router_vendor(router.owner_subscription_id)
    iptrunk_number_of_members = initial_user_input["iptrunk_number_of_members"]
    if router_vendor == Vendor.NOKIA:
        iptrunk_speed = initial_user_input["iptrunk_speed"]

        class NokiaLAGMember(LAGMember):
            interface_name: (  # type: ignore[valid-type]
                available_interfaces_choices_including_current_members(
                    router.owner_subscription_id,
                    iptrunk_speed,
                    subscription.iptrunk.iptrunk_sides[side_index].iptrunk_side_ae_members,
                )
                if iptrunk_speed == subscription.iptrunk.iptrunk_speed
                else (
                    available_interfaces_choices(
                        router.owner_subscription_id,
                        initial_user_input["iptrunk_speed"],
                    )
                )
            )

        return Annotated[
            list[NokiaLAGMember],
            AfterValidator(validate_unique_list),
            Len(min_length=iptrunk_number_of_members, max_length=iptrunk_number_of_members),
        ]  # type: ignore[return-value]

    return Annotated[
        list[LAGMember],
        AfterValidator(validate_unique_list),
        Len(min_length=iptrunk_number_of_members, max_length=iptrunk_number_of_members),
    ]  # type: ignore[return-value]


def initial_input_form_generator(subscription_id: UUIDstr) -> FormGenerator:
    """Gather input from the operator on the interfaces that should be modified."""
    subscription = Iptrunk.from_subscription(subscription_id)

    class ModifyIptrunkForm(FormPage):
        tt_number: str
        geant_s_sid: str | None = subscription.iptrunk.geant_s_sid
        iptrunk_description: str | None = subscription.iptrunk.iptrunk_description
        iptrunk_type: IptrunkType = subscription.iptrunk.iptrunk_type
        warning_label: Label = (
            "Changing the PhyPortCapacity will result in the deletion of all AE members. "
            "You will need to add the new AE members in the next steps."
        )
        iptrunk_speed: PhysicalPortCapacity = subscription.iptrunk.iptrunk_speed
        iptrunk_number_of_members: int = len(subscription.iptrunk.iptrunk_sides[0].iptrunk_side_ae_members)
        iptrunk_isis_metric: ReadOnlyField(subscription.iptrunk.iptrunk_isis_metric, default_type=int)  # type: ignore[valid-type]
        iptrunk_ipv4_network: ReadOnlyField(  # type: ignore[valid-type]
            str(subscription.iptrunk.iptrunk_ipv4_network), default_type=IPv4AddressType
        )
        iptrunk_ipv6_network: ReadOnlyField(  # type: ignore[valid-type]
            str(subscription.iptrunk.iptrunk_ipv6_network), default_type=IPv6AddressType
        )

        @field_validator("tt_number")
        def validate_tt_number(cls, tt_number: str) -> str:
            return validate_tt_number(tt_number)

    initial_user_input = yield ModifyIptrunkForm

    class VerifyMinimumLinksForm(FormPage):
        info_label: Label = (
            f"This is the calculated minimum-links for this LAG: " f"{initial_user_input.iptrunk_number_of_members - 1}"
        )
        iptrunk_minimum_links: int = initial_user_input.iptrunk_number_of_members - 1
        info_label2: Label = "Please confirm or modify."

    verify_minimum_links = yield VerifyMinimumLinksForm
    ae_members_side_a = initialize_ae_members(subscription, initial_user_input.model_dump(), 0)

    class ModifyIptrunkSideAForm(FormPage):
        model_config = ConfigDict(title="Provide subscription details for side A of the trunk.")

        side_a_node: ReadOnlyField(  # type: ignore[valid-type]
            subscription.iptrunk.iptrunk_sides[0].iptrunk_side_node.router_fqdn, default_type=str
        )
        side_a_ae_iface: ReadOnlyField(subscription.iptrunk.iptrunk_sides[0].iptrunk_side_ae_iface, default_type=str)  # type: ignore[valid-type]
        side_a_ae_geant_a_sid: str | None = subscription.iptrunk.iptrunk_sides[0].iptrunk_side_ae_geant_a_sid
        side_a_ae_members: ae_members_side_a = (  # type: ignore[valid-type]
            subscription.iptrunk.iptrunk_sides[0].iptrunk_side_ae_members
            if initial_user_input.iptrunk_speed == subscription.iptrunk.iptrunk_speed
            else []
        )

        @field_validator("side_a_ae_members")
        def validate_iptrunk_unique_interface_side_a(cls, side_a_ae_members: list[LAGMember]) -> list[LAGMember]:
            return validate_iptrunk_unique_interface(side_a_ae_members)

        @field_validator("side_a_ae_members")
        def validate_interface_name_members(cls, side_a_ae_members: list[LAGMember]) -> list[LAGMember]:
            vendor = subscription.iptrunk.iptrunk_sides[0].iptrunk_side_node.vendor
            return validate_interface_name_list(side_a_ae_members, vendor)

    user_input_side_a = yield ModifyIptrunkSideAForm
    ae_members_side_b = initialize_ae_members(subscription, initial_user_input.model_dump(), 1)

    class ModifyIptrunkSideBForm(FormPage):
        model_config = ConfigDict(title="Provide subscription details for side B of the trunk.")

        side_b_node: ReadOnlyField(  # type: ignore[valid-type]
            subscription.iptrunk.iptrunk_sides[1].iptrunk_side_node.router_fqdn, default_type=str
        )
        side_b_ae_iface: ReadOnlyField(subscription.iptrunk.iptrunk_sides[1].iptrunk_side_ae_iface, default_type=str)  # type: ignore[valid-type]
        side_b_ae_geant_a_sid: str | None = subscription.iptrunk.iptrunk_sides[1].iptrunk_side_ae_geant_a_sid
        side_b_ae_members: ae_members_side_b = (  # type: ignore[valid-type]
            subscription.iptrunk.iptrunk_sides[1].iptrunk_side_ae_members
            if initial_user_input.iptrunk_speed == subscription.iptrunk.iptrunk_speed
            else []
        )

        @field_validator("side_b_ae_members")
        def validate_iptrunk_unique_interface_side_b(cls, side_b_ae_members: list[LAGMember]) -> list[LAGMember]:
            return validate_iptrunk_unique_interface(side_b_ae_members)

        @field_validator("side_b_ae_members")
        def validate_interface_name_members(cls, side_b_ae_members: list[LAGMember]) -> list[LAGMember]:
            vendor = subscription.iptrunk.iptrunk_sides[1].iptrunk_side_node.vendor
            return validate_interface_name_list(side_b_ae_members, vendor)

    user_input_side_b = yield ModifyIptrunkSideBForm

    return (
        initial_user_input.model_dump()
        | user_input_side_a.model_dump()
        | user_input_side_b.model_dump()
        | verify_minimum_links.model_dump()
    )


@step("Determine whether we should be running interface checks")
def determine_change_in_capacity(
    subscription: Iptrunk, iptrunk_speed: str, side_a_ae_members: list[LAGMember], side_b_ae_members: list[LAGMember]
) -> State:
    """Determine whether we should run pre- and post-checks on the IP trunk.

    This can be caused by the following conditions:
     * The total capacity of the trunk changes
     * The amount of interfaces changes
     * One or more interface names have changed on side A
     * One or more interface names have changed on side B
    """
    capacity_has_changed = (
        iptrunk_speed != subscription.iptrunk.iptrunk_speed
        or len(side_a_ae_members) != len(subscription.iptrunk.iptrunk_sides[0].iptrunk_side_ae_members)
        or any(
            old_interface.interface_name != new_interface.interface_name
            for old_interface, new_interface in zip(
                subscription.iptrunk.iptrunk_sides[0].iptrunk_side_ae_members, side_a_ae_members, strict=False
            )
        )
        or any(
            old_interface.interface_name != new_interface.interface_name
            for old_interface, new_interface in zip(
                subscription.iptrunk.iptrunk_sides[1].iptrunk_side_ae_members, side_b_ae_members, strict=False
            )
        )
    )

    return {"capacity_has_changed": capacity_has_changed}


@step("Check IP connectivity of the trunk")
def check_ip_trunk_connectivity(subscription: Iptrunk, callback_route: str) -> State:
    """Check successful connectivity across a trunk."""
    extra_vars = {"wfo_ip_trunk_json": json.loads(json_dumps(subscription)), "check": "ping"}

    execute_playbook(
        playbook_name="iptrunks_checks.yaml",
        callback_route=callback_route,
        inventory=subscription.iptrunk.iptrunk_sides[0].iptrunk_side_node.router_fqdn,
        extra_vars=extra_vars,
    )

    return {"subscription": subscription}


@step("Check LLDP on the trunk endpoints")
def check_ip_trunk_lldp(subscription: Iptrunk, callback_route: str) -> State:
    """Check LLDP on trunk endpoints."""
    extra_vars = {"wfo_ip_trunk_json": json.loads(json_dumps(subscription)), "check": "lldp"}

    execute_playbook(
        playbook_name="iptrunks_checks.yaml",
        callback_route=callback_route,
        inventory=f"{subscription.iptrunk.iptrunk_sides[0].iptrunk_side_node.router_fqdn}\n"
        f"{subscription.iptrunk.iptrunk_sides[1].iptrunk_side_node.router_fqdn}\n",
        extra_vars=extra_vars,
    )

    return {"subscription": subscription}


@step("Update subscription")
def modify_iptrunk_subscription(
    subscription: Iptrunk,
    geant_s_sid: str | None,
    iptrunk_type: IptrunkType,
    iptrunk_description: str | None,
    iptrunk_speed: PhysicalPortCapacity,
    iptrunk_minimum_links: int,
    side_a_ae_geant_a_sid: str | None,
    side_a_ae_members: list[dict],
    side_b_ae_geant_a_sid: str | None,
    side_b_ae_members: list[dict],
) -> State:
    """Modify the subscription in the service database, reflecting the changes to the newly selected interfaces."""
    # Prepare the list of removed AE members
    previous_ae_members = [
        [
            {
                "interface_name": member.interface_name,
                "interface_description": member.interface_description,
            }
            for member in side.iptrunk_side_ae_members
        ]
        for side in subscription.iptrunk.iptrunk_sides
    ]
    removed_ae_members = []

    for side_index in range(2):
        previous_members = previous_ae_members[side_index]
        current_members = side_a_ae_members if side_index == 0 else side_b_ae_members
        removed_ae_members.append([ae_member for ae_member in previous_members if ae_member not in current_members])

    subscription.iptrunk.geant_s_sid = geant_s_sid
    subscription.iptrunk.iptrunk_description = iptrunk_description
    subscription.iptrunk.iptrunk_type = iptrunk_type
    subscription.iptrunk.iptrunk_speed = iptrunk_speed
    subscription.iptrunk.iptrunk_minimum_links = iptrunk_minimum_links

    subscription.iptrunk.iptrunk_sides[0].iptrunk_side_ae_geant_a_sid = side_a_ae_geant_a_sid
    #  Flush the old list of member interfaces
    subscription.iptrunk.iptrunk_sides[0].iptrunk_side_ae_members.clear()
    #  And update the list to only include the new member interfaces
    for member in side_a_ae_members:
        subscription.iptrunk.iptrunk_sides[0].iptrunk_side_ae_members.append(
            IptrunkInterfaceBlock.new(subscription_id=uuid4(), **member),
        )

    subscription.iptrunk.iptrunk_sides[1].iptrunk_side_ae_geant_a_sid = side_b_ae_geant_a_sid
    subscription.iptrunk.iptrunk_sides[1].iptrunk_side_ae_members.clear()
    for member in side_b_ae_members:
        subscription.iptrunk.iptrunk_sides[1].iptrunk_side_ae_members.append(
            IptrunkInterfaceBlock.new(subscription_id=uuid4(), **member),
        )

    side_names = sorted([
        subscription.iptrunk.iptrunk_sides[0].iptrunk_side_node.router_site.site_name,
        subscription.iptrunk.iptrunk_sides[1].iptrunk_side_node.router_site.site_name,
    ])
    subscription.description = f"IP trunk {side_names[0]} {side_names[1]}, geant_s_sid:{geant_s_sid}"

    return {
        "subscription": subscription,
        "removed_ae_members": removed_ae_members,
        "previous_ae_members": previous_ae_members,
    }


@step("[DRY RUN] Provision IP trunk interface")
def provision_ip_trunk_iface_dry(
    subscription: Iptrunk,
    process_id: UUIDstr,
    callback_route: str,
    tt_number: str,
    removed_ae_members: list[str],
) -> State:
    """Perform a dry run of deploying the updated IP trunk."""
    extra_vars = {
        "wfo_trunk_json": json.loads(json_dumps(subscription)),
        "dry_run": True,
        "verb": "deploy",
        "config_object": "trunk_interface",
        "removed_ae_members": removed_ae_members,
        "commit_comment": f"GSO_PROCESS_ID: {process_id} - TT_NUMBER: {tt_number} - Deploy config for "
        f"{subscription.iptrunk.geant_s_sid}",
    }

    execute_playbook(
        playbook_name="iptrunks.yaml",
        callback_route=callback_route,
        inventory=f"{subscription.iptrunk.iptrunk_sides[0].iptrunk_side_node.router_fqdn}\n"
        f"{subscription.iptrunk.iptrunk_sides[1].iptrunk_side_node.router_fqdn}\n",
        extra_vars=extra_vars,
    )

    return {"subscription": subscription}


@step("[FOR REAL] Provision IP trunk interface")
def provision_ip_trunk_iface_real(
    subscription: Iptrunk,
    process_id: UUIDstr,
    callback_route: str,
    tt_number: str,
    removed_ae_members: list[str],
) -> State:
    """Provision the new IP trunk with updated interfaces."""
    extra_vars = {
        "wfo_trunk_json": json.loads(json_dumps(subscription)),
        "dry_run": False,
        "verb": "deploy",
        "config_object": "trunk_interface",
        "removed_ae_members": removed_ae_members,
        "commit_comment": f"GSO_PROCESS_ID: {process_id} - TT_NUMBER: {tt_number} - Deploy config for "
        f"{subscription.iptrunk.geant_s_sid}",
    }

    execute_playbook(
        playbook_name="iptrunks.yaml",
        callback_route=callback_route,
        inventory=f"{subscription.iptrunk.iptrunk_sides[0].iptrunk_side_node.router_fqdn}\n"
        f"{subscription.iptrunk.iptrunk_sides[1].iptrunk_side_node.router_fqdn}\n",
        extra_vars=extra_vars,
    )

    return {"subscription": subscription}


def _netbox_update_interfaces(
    subscription_id: UUID,
    side_block: IptrunkSideBlock,
    removed_ae_members: list[dict],
    previous_ae_members: list[dict],
) -> None:
    nbclient = NetboxClient()

    # Free removed interfaces
    for member in removed_ae_members:
        nbclient.free_interface(side_block.iptrunk_side_node.router_fqdn, member["interface_name"])

    # Attach physical interfaces to :term:`LAG`
    # Update interface description to subscription ID
    # Reserve interfaces
    for interface in side_block.iptrunk_side_ae_members:
        if any(ae_member["interface_name"] == interface.interface_name for ae_member in previous_ae_members):
            continue
        if side_block.iptrunk_side_ae_iface:
            nbclient.attach_interface_to_lag(
                device_name=side_block.iptrunk_side_node.router_fqdn,
                lag_name=side_block.iptrunk_side_ae_iface,
                iface_name=interface.interface_name,
                description=str(subscription_id),
            )

        nbclient.reserve_interface(
            device_name=side_block.iptrunk_side_node.router_fqdn,
            iface_name=interface.interface_name,
        )


@step("Netbox: Reserve side A interfaces")
def netbox_update_interfaces_side_a(
    subscription: Iptrunk, removed_ae_members: list[list[dict]], previous_ae_members: list[list[dict]]
) -> None:
    """Update Netbox such that it contains the new interfaces on side A."""
    _netbox_update_interfaces(
        subscription.subscription_id,
        subscription.iptrunk.iptrunk_sides[0],
        removed_ae_members[0],
        previous_ae_members[0],
    )


@step("Netbox: Reserve side B interfaces")
def netbox_update_interfaces_side_b(
    subscription: Iptrunk, removed_ae_members: list[list[dict]], previous_ae_members: list[list[dict]]
) -> None:
    """Update Netbox such that it contains the new interfaces on side B."""
    _netbox_update_interfaces(
        subscription.subscription_id,
        subscription.iptrunk.iptrunk_sides[1],
        removed_ae_members[1],
        previous_ae_members[1],
    )


def _netbox_allocate_interfaces(side_block: IptrunkSideBlock, previous_ae_members: list[dict]) -> None:
    nbclient = NetboxClient()

    for interface in side_block.iptrunk_side_ae_members:
        if any(ae_member["interface_name"] == interface.interface_name for ae_member in previous_ae_members):
            continue
        nbclient.allocate_interface(
            device_name=side_block.iptrunk_side_node.router_fqdn,
            iface_name=interface.interface_name,
        )

    # detach the old interfaces from lag
    if side_block.iptrunk_side_ae_iface:
        nbclient.detach_interfaces_from_lag(
            device_name=side_block.iptrunk_side_node.router_fqdn,
            lag_name=side_block.iptrunk_side_ae_iface,
        )


@step("Netbox: Allocate side A interfaces")
def allocate_interfaces_in_netbox_side_a(subscription: Iptrunk, previous_ae_members: list[list[dict]]) -> None:
    """Allocate the :term:`LAG` interfaces on side A in Netbox.

    Attach the :term:`LAG` interface to the physical interface detach old one from the :term:`LAG`.
    """
    _netbox_allocate_interfaces(subscription.iptrunk.iptrunk_sides[0], previous_ae_members[0])


@step("Netbox: Allocate side B interfaces")
def allocate_interfaces_in_netbox_side_b(subscription: Iptrunk, previous_ae_members: list[list[dict]]) -> None:
    """Allocate the :term:`LAG` interface on side B in Netbox.

    Attach the :term:`LAG` interface to the physical interface detach old one from the :term:`LAG`.
    """
    _netbox_allocate_interfaces(subscription.iptrunk.iptrunk_sides[1], previous_ae_members[1])


@step("Check Optical POST levels on the trunk endpoint")
def check_ip_trunk_optical_levels_post(subscription: Iptrunk, callback_route: str) -> State:
    """Check Optical POST levels on the trunk."""
    extra_vars = {"wfo_ip_trunk_json": json.loads(json_dumps(subscription)), "check": "optical_post"}

    execute_playbook(
        playbook_name="iptrunks_checks.yaml",
        callback_route=callback_route,
        inventory=f"{subscription.iptrunk.iptrunk_sides[0].iptrunk_side_node.router_fqdn}\n"
        f"{subscription.iptrunk.iptrunk_sides[1].iptrunk_side_node.router_fqdn}\n",
        extra_vars=extra_vars,
    )

    return {"subscription": subscription}


@workflow(
    "Modify IP Trunk interface",
    initial_input_form=wrap_modify_initial_input_form(initial_input_form_generator),
    target=Target.MODIFY,
)
def modify_trunk_interface() -> StepList:
    """Modify the interfaces that are part of an IP trunk.

    * Update the subscription in the database
    * Reserve new interfaces in Netbox
    * Provision the updated version of the IP trunk, first as a dry run
    * Allocate the reserved interfaces in Netbox
    """
    side_a_is_nokia = conditional(
        lambda state: get_router_vendor(
            state["subscription"]["iptrunk"]["iptrunk_sides"][0]["iptrunk_side_node"]["owner_subscription_id"]
        )
        == Vendor.NOKIA
    )
    side_b_is_nokia = conditional(
        lambda state: get_router_vendor(
            state["subscription"]["iptrunk"]["iptrunk_sides"][1]["iptrunk_side_node"]["owner_subscription_id"]
        )
        == Vendor.NOKIA
    )
    capacity_has_changed = conditional(lambda state: state["capacity_has_changed"])

    return (
        begin
        >> store_process_subscription(Target.MODIFY)
        >> unsync
        >> determine_change_in_capacity
        >> capacity_has_changed(lso_interaction(check_ip_trunk_lldp))
        >> capacity_has_changed(lso_interaction(check_ip_trunk_optical_levels_pre))
        >> capacity_has_changed(lso_interaction(check_ip_trunk_connectivity))
        >> capacity_has_changed(lso_interaction(check_ip_trunk_isis))
        >> modify_iptrunk_subscription
        >> side_a_is_nokia(netbox_update_interfaces_side_a)
        >> side_b_is_nokia(netbox_update_interfaces_side_b)
        >> lso_interaction(provision_ip_trunk_iface_dry)
        >> lso_interaction(provision_ip_trunk_iface_real)
        >> side_a_is_nokia(allocate_interfaces_in_netbox_side_a)
        >> side_b_is_nokia(allocate_interfaces_in_netbox_side_b)
        >> capacity_has_changed(lso_interaction(check_ip_trunk_lldp))
        >> capacity_has_changed(lso_interaction(check_ip_trunk_optical_levels_post))
        >> capacity_has_changed(lso_interaction(check_ip_trunk_connectivity))
        >> capacity_has_changed(lso_interaction(check_ip_trunk_isis))
        >> resync
        >> done
    )
